#include "RecastNavMesh.hpp"
#include "Recast.h"
#include <godot_cpp/classes/global_constants.hpp>
#include <godot_cpp/core/object.hpp>

godot::RecastNavMesh::RecastNavMesh() : Node3D(), rcContext() {
}

godot::RecastNavMesh::~RecastNavMesh() {
  cleanup();
}

void godot::RecastNavMesh::_bind_methods() {
  // Context
  ADD_SIGNAL(MethodInfo("do_log", PropertyInfo(Variant::INT, "category"), PropertyInfo(Variant::STRING, "msg")));
  // Config
  ClassDB::bind_method(D_METHOD("get_width"), &RecastNavMesh::get_width);
  ClassDB::bind_method(D_METHOD("get_height"), &RecastNavMesh::get_height);

  ClassDB::bind_method(D_METHOD("set_tile_size", "tile_size"), &RecastNavMesh::set_tile_size);
  ClassDB::bind_method(D_METHOD("get_tile_size"), &RecastNavMesh::get_tile_size);
  ADD_PROPERTY(PropertyInfo(Variant::FLOAT, "tile_size"), "set_tile_size", "get_tile_size");

  ClassDB::bind_method(D_METHOD("set_border_size", "border_size"), &RecastNavMesh::set_border_size);
  ClassDB::bind_method(D_METHOD("get_border_size"), &RecastNavMesh::get_border_size);
  ADD_PROPERTY(PropertyInfo(Variant::FLOAT, "border_size"), "set_border_size", "get_border_size");

  ClassDB::bind_method(D_METHOD("set_cs", "cs"), &RecastNavMesh::set_cs);
  ClassDB::bind_method(D_METHOD("get_cs"), &RecastNavMesh::get_cs);
  ADD_PROPERTY(PropertyInfo(Variant::FLOAT, "cell_size"), "set_cs", "get_cs");

  ClassDB::bind_method(D_METHOD("set_ch", "ch"), &RecastNavMesh::set_ch);
  ClassDB::bind_method(D_METHOD("get_ch"), &RecastNavMesh::get_ch);
  ADD_PROPERTY(PropertyInfo(Variant::FLOAT, "cell_height"), "set_ch", "get_ch");

  ClassDB::bind_method(D_METHOD("set_bmin", "bmin"), &RecastNavMesh::set_bmin);
  ClassDB::bind_method(D_METHOD("get_bmin"), &RecastNavMesh::get_bmin);
  ADD_PROPERTY(PropertyInfo(Variant::VECTOR3, "bounds_min"), "set_bmin", "get_bmin");

  ClassDB::bind_method(D_METHOD("set_bmax", "bmax"), &RecastNavMesh::set_bmax);
  ClassDB::bind_method(D_METHOD("get_bmax"), &RecastNavMesh::get_bmax);
  ADD_PROPERTY(PropertyInfo(Variant::VECTOR3, "bounds_max"), "set_bmax", "get_bmax");

  ClassDB::bind_method(D_METHOD("set_walkable_slope_angle", "walkable_slope_angle"), &RecastNavMesh::set_walkable_slope_angle);
  ClassDB::bind_method(D_METHOD("get_walkable_slope_angle"), &RecastNavMesh::get_walkable_slope_angle);
  ADD_PROPERTY(PropertyInfo(Variant::FLOAT, "walkable_slope_angle"), "set_walkable_slope_angle", "get_walkable_slope_angle");

  ClassDB::bind_method(D_METHOD("set_walkable_height", "walkable_height"), &RecastNavMesh::set_walkable_height);
  ClassDB::bind_method(D_METHOD("get_walkable_height"), &RecastNavMesh::get_walkable_height);
  ADD_PROPERTY(PropertyInfo(Variant::INT, "walkable_height"), "set_walkable_height", "get_walkable_height");

  ClassDB::bind_method(D_METHOD("set_filter_low_hanging_obstacles", "filter_low_hanging_obstacles"), &RecastNavMesh::set_filter_low_hanging_obstacles);
  ClassDB::bind_method(D_METHOD("get_filter_low_hanging_obstacles"), &RecastNavMesh::get_filter_low_hanging_obstacles);
  ADD_PROPERTY(PropertyInfo(Variant::BOOL, "filter_low_hanging_obstacles"), "set_filter_low_hanging_obstacles", "get_filter_low_hanging_obstacles");

  ClassDB::bind_method(D_METHOD("set_filter_ledge_spans", "filter_ledge_spans"), &RecastNavMesh::set_filter_ledge_spans);
  ClassDB::bind_method(D_METHOD("get_filter_ledge_spans"), &RecastNavMesh::get_filter_ledge_spans);
  ADD_PROPERTY(PropertyInfo(Variant::BOOL, "filter_ledge_spans"), "set_filter_ledge_spans", "get_filter_ledge_spans");

  ClassDB::bind_method(D_METHOD("set_filter_walkable_low_height_spans", "filter_walkable_low_height_spans"), &RecastNavMesh::set_filter_walkable_low_height_spans);
  ClassDB::bind_method(D_METHOD("get_filter_walkable_low_height_spans"), &RecastNavMesh::get_filter_walkable_low_height_spans);
  ADD_PROPERTY(PropertyInfo(Variant::BOOL, "filter_walkable_low_height_spans"), "set_filter_walkable_low_height_spans", "get_filter_walkable_low_height_spans");

  ClassDB::bind_method(D_METHOD("set_walkable_climb", "walkable_climb"), &RecastNavMesh::set_walkable_climb);
  ClassDB::bind_method(D_METHOD("get_walkable_climb"), &RecastNavMesh::get_walkable_climb);
  ADD_PROPERTY(PropertyInfo(Variant::INT, "walkable_climb"), "set_walkable_climb", "get_walkable_climb");

  ClassDB::bind_method(D_METHOD("set_walkable_radius", "walkable_radius"), &RecastNavMesh::set_walkable_radius);
  ClassDB::bind_method(D_METHOD("get_walkable_radius"), &RecastNavMesh::get_walkable_radius);
  ADD_PROPERTY(PropertyInfo(Variant::INT, "walkable_radius"), "set_walkable_radius", "get_walkable_radius");

  ClassDB::bind_method(D_METHOD("set_max_edge_len", "max_edge_len"), &RecastNavMesh::set_max_edge_len);
  ClassDB::bind_method(D_METHOD("get_max_edge_len"), &RecastNavMesh::get_max_edge_len);
  ADD_PROPERTY(PropertyInfo(Variant::INT, "max_edge_length"), "set_max_edge_len", "get_max_edge_len");

  ClassDB::bind_method(D_METHOD("set_max_simplification_error", "max_simplification_error"), &RecastNavMesh::set_max_simplification_error);
  ClassDB::bind_method(D_METHOD("get_max_simplification_error"), &RecastNavMesh::get_max_simplification_error);
  ADD_PROPERTY(PropertyInfo(Variant::FLOAT, "max_simplification_error"), "set_max_simplification_error", "get_max_simplification_error");

  ClassDB::bind_method(D_METHOD("set_min_region_area", "min_region_area"), &RecastNavMesh::set_min_region_area);
  ClassDB::bind_method(D_METHOD("get_min_region_area"), &RecastNavMesh::get_min_region_area);
  ADD_PROPERTY(PropertyInfo(Variant::INT, "min_region_area"), "set_min_region_area", "get_min_region_area");

  ClassDB::bind_method(D_METHOD("set_merge_region_area", "merge_region_area"), &RecastNavMesh::set_merge_region_area);
  ClassDB::bind_method(D_METHOD("get_merge_region_area"), &RecastNavMesh::get_merge_region_area);
  ADD_PROPERTY(PropertyInfo(Variant::INT, "merge_region_area"), "set_merge_region_area", "get_merge_region_area");

  ClassDB::bind_method(D_METHOD("set_max_verts_per_poly", "max_verts_per_poly"), &RecastNavMesh::set_max_verts_per_poly);
  ClassDB::bind_method(D_METHOD("get_max_verts_per_poly"), &RecastNavMesh::get_max_verts_per_poly);
  ADD_PROPERTY(PropertyInfo(Variant::INT, "max_verts_per_poly"), "set_max_verts_per_poly", "get_max_verts_per_poly");

  ClassDB::bind_method(D_METHOD("set_detail_sample_dist", "detail_sample_dist"), &RecastNavMesh::set_detail_sample_dist);
  ClassDB::bind_method(D_METHOD("get_detail_sample_dist"), &RecastNavMesh::get_detail_sample_dist);
  ADD_PROPERTY(PropertyInfo(Variant::FLOAT, "detail_sample_dist"), "set_detail_sample_dist", "get_detail_sample_dist");

  ClassDB::bind_method(D_METHOD("set_detail_sample_max_error", "detail_sample_max_error"), &RecastNavMesh::set_detail_sample_max_error);
  ClassDB::bind_method(D_METHOD("get_detail_sample_max_error"), &RecastNavMesh::get_detail_sample_max_error);
  ADD_PROPERTY(PropertyInfo(Variant::FLOAT, "detail_sample_max_error"), "set_detail_sample_max_error", "get_detail_sample_max_error");

  ClassDB::bind_method(D_METHOD("set_partition_type", "partition_type"), &RecastNavMesh::set_partition_type);
  ClassDB::bind_method(D_METHOD("get_partition_type"), &RecastNavMesh::get_partition_type);
  ADD_PROPERTY(PropertyInfo(Variant::INT, "partition_type", PROPERTY_HINT_ENUM, "Watershed,Monotone,Layers"), "set_partition_type", "get_partition_type");

  // Recast calls
  ClassDB::bind_method(D_METHOD("get_walkable_triangles", "vertices", "indices", "areas"), &RecastNavMesh::get_walkable_triangles);
  ClassDB::bind_method(D_METHOD("add_vertices", "vertices", "area_id"), &RecastNavMesh::add_vertices);
  ClassDB::bind_method(D_METHOD("build"), &RecastNavMesh::build);
  ClassDB::bind_method(D_METHOD("is_calculated"), &RecastNavMesh::is_calculated);
}

bool godot::RecastNavMesh::init() {
  cleanup();
  rcCalcGridSize(config.bmin, config.bmax, config.cs, &config.width, &config.height);
  m_heightfield = rcAllocHeightfield();
  if(!m_heightfield) {
    log(RC_LOG_ERROR, "couldn't alloc heightfield!");
    return false;
  }
  if(!rcCreateHeightfield(this, *m_heightfield, config.width, config.height, config.bmin, config.bmax, config.cs, config.ch)) {
    log(RC_LOG_ERROR, "couldn't build heightfield!");
    return false;
  }
  //
  return false;
}

void godot::RecastNavMesh::cleanup() {

  if(m_heightfield) {
    rcFreeHeightField(m_heightfield);
    m_heightfield = NULL;
  }
  if(m_compact_heightfield) {
    rcFreeCompactHeightfield(m_compact_heightfield);
    m_compact_heightfield = NULL;
  }
  if(m_contour_set) {
    rcFreeContourSet(m_contour_set);
    m_contour_set = NULL;
  }
  if(m_poly_mesh) {
    rcFreePolyMesh(m_poly_mesh);
    m_poly_mesh = NULL;
  }
  if(m_poly_mesh_detail) {
    rcFreePolyMeshDetail(m_poly_mesh_detail);
    m_poly_mesh_detail = NULL;
  }
}

void godot::RecastNavMesh::get_walkable_triangles(PackedFloat32Array vertices, PackedInt32Array indices, PackedByteArray areas) {
  areas.resize(indices.size() / 3);
  areas.fill(0);
  rcMarkWalkableTriangles(this, config.walkableSlopeAngle, vertices.ptr(), vertices.size() / 3, indices.ptr(), indices.size() / 3, areas.ptrw());
}

bool godot::RecastNavMesh::add_vertices(PackedFloat32Array vertices, PackedInt32Array indices, PackedByteArray areas) {
  bool ret = rcRasterizeTriangles(this, vertices.ptr(), vertices.size() / 3, indices.ptr(), areas.ptr(), indices.size() / 3, *m_heightfield, config.walkableClimb);
  if(!ret) {
    log(RC_LOG_ERROR, "couldn't rasterize triangles!");
  }
  return ret;
}

bool godot::RecastNavMesh::build() {
  if(this->get_filter_low_hanging_obstacles()) {
    rcFilterLowHangingWalkableObstacles(this, config.walkableClimb, *m_heightfield);
  }
  if(this->get_filter_ledge_spans()) {
    rcFilterLedgeSpans(this, config.walkableHeight, config.walkableClimb, *m_heightfield);
  }
  if(this->get_filter_walkable_low_height_spans()) {
    rcFilterWalkableLowHeightSpans(this, config.walkableHeight, *m_heightfield);
  }
  m_compact_heightfield = rcAllocCompactHeightfield();

  if(!m_compact_heightfield) {
    log(RC_LOG_ERROR, "couldn't alloc compact heightfield!");
    return false;
  }

  if(!rcBuildCompactHeightfield(this, config.walkableHeight, config.walkableClimb, *m_heightfield, *m_compact_heightfield)) {
    log(RC_LOG_ERROR, "couldn't build compact heightfield!");
    return false;
  }

  rcFreeHeightField(m_heightfield);
  m_heightfield = NULL;

  if (!rcErodeWalkableArea(this, config.walkableRadius, *m_compact_heightfield))
  {
    log(RC_LOG_ERROR, "buildNavigation: Could not erode.");
    return false;
  }

  if (partition_type == WATERSHED)
  {
    // Prepare for region partitioning, by calculating distance field along the walkable surface.
    if (!rcBuildDistanceField(this, *m_compact_heightfield))
    {
      log(RC_LOG_ERROR, "buildNavigation: Could not build distance field.");
      return false;
    }

    // Partition the walkable surface into simple regions without holes.
    if (!rcBuildRegions(this, *m_compact_heightfield, 0, config.minRegionArea, config.mergeRegionArea))
    {
      log(RC_LOG_ERROR, "buildNavigation: Could not build watershed regions.");
      return false;
    }
  }
  else if (partition_type == MONOTONE)
  {
    // Partition the walkable surface into simple regions without holes.
    // Monotone partitioning does not need distancefield.
    if (!rcBuildRegionsMonotone(this, *m_compact_heightfield, 0, config.minRegionArea, config.mergeRegionArea))
    {
      log(RC_LOG_ERROR, "buildNavigation: Could not build monotone regions.");
      return false;
    }
  }
  else // SAMPLE_PARTITION_LAYERS
  {
    // Partition the walkable surface into simple regions without holes.
    if (!rcBuildLayerRegions(this, *m_compact_heightfield, 0, config.minRegionArea))
    {
      log(RC_LOG_ERROR, "buildNavigation: Could not build layer regions.");
      return false;
    }
  }

  m_contour_set = rcAllocContourSet();
  if(!m_contour_set) {
    log(RC_LOG_ERROR, "couldn't alloc contour set!");
    return false;
  }

  if(!rcBuildContours(this, *m_compact_heightfield, config.maxSimplificationError, config.maxEdgeLen, *m_contour_set)) {
    log(RC_LOG_ERROR, "couldn't build contours!");
    return false;
  }

  m_poly_mesh = rcAllocPolyMesh();
  if(!m_poly_mesh) {
    log(RC_LOG_ERROR, "couldn't alloc poly mesh!");
    return false;
  }

  if(!rcBuildPolyMesh(this, *m_contour_set, config.maxVertsPerPoly, *m_poly_mesh)) {
    log(RC_LOG_ERROR, "couldn't build poly mesh!");
    return false;
  }

  m_poly_mesh_detail = rcAllocPolyMeshDetail();
  if(!m_poly_mesh_detail) {
    log(RC_LOG_ERROR, "couldn't alloc detail mesh!");
    return false;
  }

  if(!rcBuildPolyMeshDetail(this, *m_poly_mesh, *m_compact_heightfield, config.detailSampleDist, config.detailSampleMaxError, *m_poly_mesh_detail)) {
    log(RC_LOG_ERROR, "couldn't build detail mesh!");
    return false;
  }

  rcFreeCompactHeightfield(m_compact_heightfield);
  m_compact_heightfield = NULL;
  rcFreeContourSet(m_contour_set);
  m_contour_set = NULL;

  m_calculated = true;
  return true;
}

